@@ -124,7 +124,7 @@ void do_open_filesys(int argc, char **argv)
 	int	catastrophic = 0;
 	blk_t	superblock = 0;
 	blk_t	blocksize = 0;
-	int	open_flags = 0;
+	int	open_flags = EXT2_FLAG_SOFTSUPP_FEATURES;
 	char	*data_filename = 0;
 	
 	reset_getopt();
@@ -181,9 +181,10 @@ print_usage:
 
 void do_lcd(int argc, char **argv)
 {
-	if (common_args_process(argc, argv, 2, 2, "lcd",
-				"<native dir>", 0))
+	if (argc != 2) {
+		com_err(argv[0], 0, "Usage: %s %s", argv[0], "<native dir>");
 		return;
+	}
 
 	if (chdir(argv[1]) == -1) {
 		com_err(argv[0], errno,
@@ -266,13 +267,26 @@ static void print_features(struct ext2_super_block * s, FILE *f)
 	fputs("\n", f);
 }
 
+static void print_bg_opts(struct ext2_group_desc *gdp, int mask,
+			  const char *str, int *first, FILE *f)
+{
+	if (gdp->bg_flags & mask) {
+		if (*first) {
+			fputs("           [", f);
+			*first = 0;
+		} else
+			fputs(", ", f);
+		fputs(str, f);
+	}
+}
+
 void do_show_super_stats(int argc, char *argv[])
 {
 	dgrp_t	i;
 	FILE 	*out;
 	struct ext2_group_desc *gdp;
 	int	c, header_only = 0;
-	int	numdirs = 0;
+	int	numdirs = 0, first;
 
 	reset_getopt();
 	while ((c = getopt (argc, argv, "h")) != EOF) {
@@ -302,7 +316,7 @@ void do_show_super_stats(int argc, char *argv[])
 	}
 	
 	gdp = &current_fs->group_desc[0];
-	for (i = 0; i < current_fs->group_desc_count; i++, gdp++)
+	for (i = 0; i < current_fs->group_desc_count; i++, gdp++) {
 		fprintf(out, " Group %2d: block bitmap at %u, "
 		        "inode bitmap at %u, "
 		        "inode table at %u\n"
@@ -318,6 +332,14 @@ void do_show_super_stats(int argc, char *argv[])
 		        gdp->bg_used_dirs_count,
 		        gdp->bg_used_dirs_count != 1 ? "directories"
 				: "directory");
+		first = 1;
+		print_bg_opts(gdp, EXT2_BG_INODE_UNINIT, "Inode not init",
+			      &first, out);
+		print_bg_opts(gdp, EXT2_BG_BLOCK_UNINIT, "Block not init",
+			      &first, out);
+		if (!first)
+			fputs("]\n", out);
+	}
 	close_pager(out);
 	return;
 print_usage:
@@ -356,10 +378,12 @@ static void finish_range(struct list_blocks_struct *lb)
 	else
 		fprintf(lb->f, ", ");
 	if (lb->first_block == lb->last_block)
-		fprintf(lb->f, "(%lld):%u", lb->first_bcnt, lb->first_block);
+		fprintf(lb->f, "(%lld):%u",
+			(long long)lb->first_bcnt, lb->first_block);
 	else
-		fprintf(lb->f, "(%lld-%lld):%u-%u", lb->first_bcnt,
-			lb->last_bcnt, lb->first_block, lb->last_block);
+		fprintf(lb->f, "(%lld-%lld):%u-%u",
+			(long long)lb->first_bcnt, (long long)lb->last_bcnt,
+			lb->first_block, lb->last_block);
 	lb->first_block = 0;
 }
 
@@ -422,13 +446,14 @@ static void dump_xattr_string(FILE *out, const char *str, int len)
 
 	for (i = 0; i < len; i++)
 		if (printable)
-			fprintf(out, "%c", str[i]);
+			fprintf(out, "%c", (unsigned char)str[i]);
 		else
-			fprintf(out, "%02x ", str[i]);
+			fprintf(out, "%02x ", (unsigned char)str[i]);
 }
 
-static void internal_dump_inode_extra(FILE *out, const char *prefix,
-				      ext2_ino_t inode_num, 
+static void internal_dump_inode_extra(FILE *out, 
+				      const char *prefix EXT2FS_ATTR((unused)),
+				      ext2_ino_t inode_num EXT2FS_ATTR((unused)), 
 				      struct ext2_inode_large *inode)
 {
 	struct ext2_ext_attr_entry *entry;
@@ -486,7 +511,7 @@ static void dump_blocks(FILE *f, const char *prefix, ext2_ino_t inode)
 			     list_blocks_proc, (void *)&lb);
 	finish_range(&lb);
 	if (lb.total)
-		fprintf(f, "\n%sTOTAL: %lld\n", prefix, lb.total);
+		fprintf(f, "\n%sTOTAL: %lld\n", prefix, (long long)lb.total);
 	fprintf(f,"\n");
 }
 
@@ -512,15 +537,15 @@ void internal_dump_inode(FILE *out, const char *prefix,
 		prefix, 
 		inode->i_mode & 0777, inode->i_flags, inode->i_generation);
 	fprintf(out, "%sUser: %5d   Group: %5d   Size: ",
-		prefix, inode->i_uid, inode->i_gid);
+		prefix, inode_uid(*inode), inode_gid(*inode));
 	if (LINUX_S_ISREG(inode->i_mode)) {
-		__u64 i_size = (inode->i_size |
-				((unsigned long long)inode->i_size_high << 32));
+		unsigned long long i_size = (inode->i_size |
+				    ((unsigned long long)inode->i_size_high << 32));
 
-		fprintf(out, "%lld\n", i_size);
+		fprintf(out, "%llu\n", i_size);
 	} else
 		fprintf(out, "%d\n", inode->i_size);
-	if (current_fs->super->s_creator_os == EXT2_OS_HURD)
+	if (os == EXT2_OS_HURD)
 		fprintf(out,
 			"%sFile ACL: %d    Directory ACL: %d Translator: %d\n",
 			prefix,
@@ -530,13 +555,16 @@ void internal_dump_inode(FILE *out, const char *prefix,
 		fprintf(out, "%sFile ACL: %d    Directory ACL: %d\n",
 			prefix,
 			inode->i_file_acl, LINUX_S_ISDIR(inode->i_mode) ? inode->i_dir_acl : 0);
-	fprintf(out, "%sLinks: %d   Blockcount: %u\n",
-		prefix, inode->i_links_count, inode->i_blocks);
+	if (os == EXT2_OS_LINUX) 
+		fprintf(out, "%sLinks: %d   Blockcount: %llu\n",
+			prefix, inode->i_links_count, 
+			(((unsigned long long) 
+			  inode->osd2.linux2.l_i_blocks_hi << 32)) + 
+			inode->i_blocks);
+	else
+		fprintf(out, "%sLinks: %d   Blockcount: %u\n",
+			prefix, inode->i_links_count, inode->i_blocks);
 	switch (os) {
-	    case EXT2_OS_LINUX:
-		frag = inode->osd2.linux2.l_i_frag;
-		fsize = inode->osd2.linux2.l_i_fsize;
-		break;
 	    case EXT2_OS_HURD:
 		frag = inode->osd2.hurd2.h_i_frag;
 		fsize = inode->osd2.hurd2.h_i_fsize;
@@ -701,7 +729,7 @@ void do_testi(int argc, char *argv[])
 void do_freeb(int argc, char *argv[])
 {
 	blk_t block;
-	int count = 1;
+	blk_t count = 1;
 
 	if (common_block_args_process(argc, argv, &block, &count))
 		return;
@@ -720,7 +748,7 @@ void do_freeb(int argc, char *argv[])
 void do_setb(int argc, char *argv[])
 {
 	blk_t block;
-	int count = 1;
+	blk_t count = 1;
 
 	if (common_block_args_process(argc, argv, &block, &count))
 		return;
@@ -739,7 +767,7 @@ void do_setb(int argc, char *argv[])
 void do_testb(int argc, char *argv[])
 {
 	blk_t block;
-	int count = 1;
+	blk_t count = 1;
 
 	if (common_block_args_process(argc, argv, &block, &count))
 		return;
@@ -846,6 +874,9 @@ void do_modify_inode(int argc, char *argv[])
 	modify_u32(argv[0], "Access time", decimal_format, &inode.i_atime);
 	modify_u32(argv[0], "Deletion time", decimal_format, &inode.i_dtime);
 	modify_u16(argv[0], "Link count", decimal_format, &inode.i_links_count);
+	if (os == EXT2_OS_LINUX)
+		modify_u16(argv[0], "Block count high", unsignedlong_format, 
+			   &inode.osd2.linux2.l_i_blocks_hi);
 	modify_u32(argv[0], "Block count", unsignedlong_format, &inode.i_blocks);
 	modify_u32(argv[0], "File flags", hex_format, &inode.i_flags);
 	modify_u32(argv[0], "Generation", hex_format, &inode.i_generation);
@@ -858,16 +889,12 @@ void do_modify_inode(int argc, char *argv[])
 	else
 		modify_u32(argv[0], "High 32bits of size", decimal_format, &inode.i_size_high);
 
-	if (current_fs->super->s_creator_os == EXT2_OS_HURD)
+	if (os == EXT2_OS_HURD)
 		modify_u32(argv[0], "Translator Block",
 			    decimal_format, &inode.osd1.hurd1.h_i_translator);
 	
 	modify_u32(argv[0], "Fragment address", decimal_format, &inode.i_faddr);
 	switch (os) {
-	    case EXT2_OS_LINUX:
-		frag = &inode.osd2.linux2.l_i_frag;
-		fsize = &inode.osd2.linux2.l_i_fsize;
-		break;
 	    case EXT2_OS_HURD:
 		frag = &inode.osd2.hurd2.h_i_frag;
 		fsize = &inode.osd2.hurd2.h_i_fsize;
@@ -929,15 +956,23 @@ void do_print_working_directory(int argc, char *argv[])
 		com_err(argv[0], retval,
 			"while trying to get pathname of cwd");
 	}
-	printf("[pwd]   INODE: %6u  PATH: %s\n", cwd, pathname);
-	free(pathname);
+	printf("[pwd]   INODE: %6u  PATH: %s\n",
+	       cwd, pathname ? pathname : "NULL");
+        if (pathname) {
+		free(pathname);
+		pathname = NULL;
+        }
 	retval = ext2fs_get_pathname(current_fs, root, 0, &pathname);
 	if (retval) {
 		com_err(argv[0], retval,
 			"while trying to get pathname of root");
 	}
-	printf("[root]  INODE: %6u  PATH: %s\n", root, pathname);
-	free(pathname);
+	printf("[root]  INODE: %6u  PATH: %s\n",
+	       root, pathname ? pathname : "NULL");
+	if (pathname) {
+		free(pathname);
+		pathname = NULL;
+	}
 	return;
 }
 
@@ -1769,7 +1804,7 @@ int main(int argc, char **argv)
 	int		sci_idx;
 	const char	*usage = "Usage: debugfs [-b blocksize] [-s superblock] [-f cmd_file] [-R request] [-V] [[-w] [-c] device]";
 	int		c;
-	int		open_flags = 0;
+	int		open_flags = EXT2_FLAG_SOFTSUPP_FEATURES;
 	char		*request = 0;
 	int		exit_status = 0;
 	char		*cmd_file = 0;
@@ -1778,7 +1813,7 @@ int main(int argc, char **argv)
 	int		catastrophic = 0;
 	char		*data_filename = 0;
 	
-	initialize_ext2_error_table();
+	add_error_table(&et_ext2_error_table);
 	fprintf (stderr, "debugfs %s (%s)\n", E2FSPROGS_VERSION,
 		 E2FSPROGS_DATE);
 
@@ -1854,5 +1889,6 @@ int main(int argc, char **argv)
 	if (current_fs)
 		close_filesystem();
 	
+	remove_error_table(&et_ext2_error_table);
 	return exit_status;
 }
