@@ -23,7 +23,7 @@
 struct deleted_info {
 	ext2_ino_t	ino;
 	unsigned short	mode;
-	unsigned short	uid;
+	__u32		uid;
 	__u64		size;
 	time_t		dtime;
 	int		num_blocks;
@@ -49,7 +49,7 @@ static int deleted_info_compare(const void *a, const void *b)
 
 static int lsdel_proc(ext2_filsys fs,
 		      blk_t	*block_nr,
-		      int blockcnt,
+		      int blockcnt EXT2FS_ATTR((unused)),
 		      void *private)
 {
 	struct lsdel_struct *lsd = (struct lsdel_struct *) private;
@@ -81,12 +81,13 @@ void do_lsdel(int argc, char **argv)
 	int			i;
  	long			secs = 0;
  	char			*tmp;
-	time_t			now = time(0);
+	time_t			now;
 	FILE			*out;
 	
 	if (common_args_process(argc, argv, 1, 2, "ls_deleted_inodes",
 				"[secs]", 0))
 		return;
+
 	if (argc > 1) {
 		secs = strtol(argv[1],&tmp,0);
 		if (*tmp) {
@@ -95,6 +96,7 @@ void do_lsdel(int argc, char **argv)
 		}
 	}
 
+	now = current_fs->now ? current_fs->now : time(0);
 	max_delarray = 100;
 	num_delarray = 0;
 	delarray = malloc(max_delarray * sizeof(struct deleted_info));
@@ -128,7 +130,7 @@ void do_lsdel(int argc, char **argv)
 	
 	while (ino) {
 		if ((inode.i_dtime == 0) ||
-		    (secs && ((now - secs) > inode.i_dtime)))
+		    (secs && ((unsigned) abs(now - secs) > inode.i_dtime)))
 			goto next;
 
 		lsd.inode = ino;
@@ -158,7 +160,7 @@ void do_lsdel(int argc, char **argv)
 				
 			delarray[num_delarray].ino = ino;
 			delarray[num_delarray].mode = inode.i_mode;
-			delarray[num_delarray].uid = inode.i_uid;
+			delarray[num_delarray].uid = inode_uid(inode);
 			delarray[num_delarray].size = inode.i_size;
 			if (!LINUX_S_ISDIR(inode.i_mode))
 				delarray[num_delarray].size |= 
