@@ -31,6 +31,7 @@
 
 static struct ext2_super_block set_sb;
 static struct ext2_inode set_inode;
+static struct ext2_group_desc set_gd;
 static ext2_ino_t set_ino;
 static int array_idx;
 
@@ -110,6 +111,7 @@ static struct field_set_info super_fields[] = {
 	{ "mkfs_time", &set_sb.s_mkfs_time, 4, parse_time },
 	{ "jnl_blocks", &set_sb.s_jnl_blocks[0], 4, parse_uint, FLAG_ARRAY, 
 	  17 },
+	{ "flags", &set_sb.s_flags, 4, parse_uint },
 	{ 0, 0, 0, 0 }
 };
 
@@ -126,6 +128,7 @@ static struct field_set_info inode_fields[] = {
 	{ "links_count", &set_inode.i_links_count, 2, parse_uint },
 	{ "blocks", &set_inode.i_blocks, 4, parse_uint },
 	{ "flags", &set_inode.i_flags, 4, parse_uint },
+	{ "version", &set_inode.osd1.linux1.l_i_version, 4, parse_uint },
 	{ "translator", &set_inode.osd1.hurd1.h_i_translator, 4, parse_uint },
 	{ "block", &set_inode.i_block[0], 4, parse_uint, FLAG_ARRAY, 
 	  EXT2_NDIR_BLOCKS },
@@ -136,8 +139,9 @@ static struct field_set_info inode_fields[] = {
 	{ "file_acl", &set_inode.i_file_acl, 4, parse_uint },
 	{ "dir_acl", &set_inode.i_dir_acl, 4, parse_uint },
 	{ "faddr", &set_inode.i_faddr, 4, parse_uint },
-	{ "frag", &set_inode.osd2.linux2.l_i_frag, 1, parse_uint },
-	{ "fsize", &set_inode.osd2.linux2.l_i_fsize, 1, parse_uint },
+	{ "blocks_hi", &set_inode.osd2.linux2.l_i_blocks_hi, 2, parse_uint },
+	{ "frag", &set_inode.osd2.hurd2.h_i_frag, 1, parse_uint },
+	{ "fsize", &set_inode.osd2.hurd2.h_i_fsize, 1, parse_uint },
 	{ "uid_high", &set_inode.osd2.linux2.l_i_uid_high, 2, parse_uint },
 	{ "gid_high", &set_inode.osd2.linux2.l_i_gid_high, 2, parse_uint },
 	{ "author", &set_inode.osd2.hurd2.h_i_author, 4, parse_uint },
@@ -145,19 +149,38 @@ static struct field_set_info inode_fields[] = {
 	{ 0, 0, 0, 0 }
 };
 
+static struct field_set_info ext2_bg_fields[] = {
+	{ "block_bitmap", &set_gd.bg_block_bitmap, 4, parse_uint },
+	{ "inode_bitmap", &set_gd.bg_inode_bitmap, 4, parse_uint },
+	{ "inode_table", &set_gd.bg_inode_table, 4, parse_uint },
+	{ "free_blocks_count", &set_gd.bg_free_blocks_count, 2, parse_uint },
+	{ "free_inodes_count", &set_gd.bg_free_inodes_count, 2, parse_uint },
+	{ "used_dirs_count", &set_gd.bg_used_dirs_count, 2, parse_uint },
+	{ "flags", &set_gd.bg_flags, 2, parse_uint },
+	{ "reserved", &set_gd.bg_reserved, 2, parse_uint, FLAG_ARRAY, 2 },
+	{ "itable_unused", &set_gd.bg_itable_unused, 2, parse_uint },
+	{ "checksum", &set_gd.bg_checksum, 2, parse_uint },
+	{ 0, 0, 0, 0 }
+};
+
+
 static struct field_set_info *find_field(struct field_set_info *fields,
 					 char *field)
 {
 	struct field_set_info *ss;
 	const char	*prefix;
 	char		*arg, *delim, *idx, *tmp;
+	int		prefix_len;
 
 	if (fields == super_fields)
 		prefix = "s_";
-	else
+	else if (fields == inode_fields)
 		prefix = "i_";
-	if (strncmp(field, prefix, 2) == 0)
-		field += 2;
+	else
+		prefix = "bg_";
+	prefix_len = strlen(prefix);
+	if (strncmp(field, prefix, prefix_len) == 0)
+		field += prefix_len;
 
 	arg = malloc(strlen(field)+1);
 	if (!arg)
@@ -278,7 +301,8 @@ static errcode_t parse_string(struct field_set_info *info, char *arg)
 
 static errcode_t parse_time(struct field_set_info *info, char *arg)
 {
-	__u32		*ptr32, t;
+	time_t		t;
+	__u32		*ptr32;
 
 	ptr32 = (__u32 *) info->ptr;
 
@@ -359,9 +383,12 @@ static void print_possible_fields(struct field_set_info *fields)
 	if (fields == super_fields) {
 		type = "Superblock";
 		cmd = "set_super_value";
-	} else {
+	} else if (fields == inode_fields) {
 		type = "Inode";
 		cmd = "set_inode";
+	} else {
+		type = "Block group descriptor";
+		cmd = "set_block_group";
 	}
 	f = open_pager();
 
@@ -457,3 +484,46 @@ void do_set_inode(int argc, char *argv[])
 			return;
 	}
 }
+
+void do_set_block_group_descriptor(int argc, char *argv[])
+{
+	const char *usage = "<bg number> <field> <value>\n"
+		"\t\"set_block_group_descriptor -l\" will list the names of "
+		"the fields in a block group descriptor\n\twhich can be set.";
+	struct field_set_info	*ss;
+	dgrp_t			set_bg;
+	char			*end;
+
+	if ((argc == 2) && !strcmp(argv[1], "-l")) {
+		print_possible_fields(ext2_bg_fields);
+		return;
+	}
+
+	if (common_args_process(argc, argv, 4, 4, "set_block_group_descriptor",
+				usage, CHECK_FS_RW))
+		return;
+
+	set_bg = strtoul(argv[1], &end, 0);
+	if (*end) {
+		com_err(argv[0], 0, "invalid block group number: %s", argv[1]);
+		return;
+	}
+
+	if (set_bg >= current_fs->group_desc_count) {
+		com_err(argv[0], 0, "block group number too big: %d", set_bg);
+		return;
+	}
+
+
+	if ((ss = find_field(ext2_bg_fields, argv[2])) == 0) {
+		com_err(argv[0], 0, "invalid field specifier: %s", argv[2]);
+		return;
+	}
+
+	set_gd = current_fs->group_desc[set_bg];
+
+	if (ss->func(ss, argv[3]) == 0) {
+		current_fs->group_desc[set_bg] = set_gd;
+		ext2fs_mark_super_dirty(current_fs);
+	}
+}
