@@ -35,6 +35,7 @@
  * 	%Id	<inode> -> i_dir_acl
  * 	%Iu	<inode> -> i_uid
  * 	%Ig	<inode> -> i_gid
+ *	%It	<inode type>
  * 	%j	<ino2>			inode number
  * 	%m	<com_err error message>
  * 	%N	<num>
@@ -70,6 +71,8 @@
  * 	@j	journal
  * 	@l	lost+found
  * 	@L	is a link
+ *	@m	multiply-claimed
+ *	@n	invalid
  * 	@o	orphaned
  * 	@p	problem in
  * 	@r	root inode
@@ -122,6 +125,8 @@ static const char *abbrevs[] = {
 	N_("hHTREE @d @i"),
 	N_("llost+found"),
 	N_("Lis a link"),
+	N_("mmultiply-claimed"),
+	N_("ninvalid"),
 	N_("oorphaned"),
 	N_("pproblem in"),
 	N_("rroot @i"),
@@ -209,7 +214,7 @@ static void print_pathname(ext2_filsys fs, ext2_ino_t dir, ext2_ino_t ino)
  */
 static _INLINE_ void expand_at_expression(e2fsck_t ctx, char ch,
 					  struct problem_context *pctx,
-					  int *first)
+					  int *first, int recurse)
 {
 	const char **cpp, *str;
 	
@@ -218,13 +223,13 @@ static _INLINE_ void expand_at_expression(e2fsck_t ctx, char ch,
 		if (ch == *cpp[0])
 			break;
 	}
-	if (*cpp) {
+	if (*cpp && recurse < 10) {
 		str = _(*cpp) + 1;
 		if (*first && islower(*str)) {
 			*first = 0;
 			fputc(toupper(*str++), stdout);
 		}
-		print_e2fsck_message(ctx, str, pctx, *first);
+		print_e2fsck_message(ctx, str, pctx, *first, recurse+1);
 	} else
 		printf("@%c", ch);
 }
@@ -237,7 +242,7 @@ static _INLINE_ void expand_inode_expression(char ch,
 {
 	struct ext2_inode	*inode;
 	struct ext2_inode_large	*large_inode;
-	char *			time_str;
+	const char *		time_str;
 	time_t			t;
 	int			do_gmt = -1;
 
@@ -259,8 +264,8 @@ static _INLINE_ void expand_inode_expression(char ch,
 			else
 				printf("%u", inode->i_size);
 #else
-			printf("%llu", (inode->i_size | 
-					((__u64) inode->i_size_high << 32)));
+			printf("%llu", inode->i_size |
+				       ((long long)inode->i_size_high << 32));
 #endif
 		}
 		break;
@@ -299,12 +304,29 @@ static _INLINE_ void expand_inode_expression(char ch,
 			      inode->i_dir_acl : 0));
 		break;
 	case 'u':
-		printf("%d", (inode->i_uid |
-			      (inode->osd2.linux2.l_i_uid_high << 16)));
+		printf("%d", inode_uid(*inode));
 		break;
 	case 'g':
-		printf("%d", (inode->i_gid |
-			      (inode->osd2.linux2.l_i_gid_high << 16)));
+		printf("%d", inode_gid(*inode));
+		break;
+	case 't':
+		if (LINUX_S_ISREG(inode->i_mode)) 
+			printf(_("regular file"));
+		else if (LINUX_S_ISDIR(inode->i_mode)) 
+			printf(_("directory"));
+		else if (LINUX_S_ISCHR(inode->i_mode)) 
+			printf(_("character device"));
+		else if (LINUX_S_ISBLK(inode->i_mode)) 
+			printf(_("block device"));
+		else if (LINUX_S_ISFIFO(inode->i_mode)) 
+			printf(_("named pipe"));
+		else if (LINUX_S_ISLNK(inode->i_mode)) 
+			printf(_("symbolic link"));
+		else if (LINUX_S_ISSOCK(inode->i_mode))
+			printf(_("socket"));
+		else
+			printf(_("unknown file type with mode 0%o"),
+			       inode->i_mode);
 		break;
 	default:
 	no_inode:
@@ -372,7 +394,7 @@ static _INLINE_ void expand_percent_expression(ext2_filsys fs, char ch,
 #ifdef EXT2_NO_64_TYPE
 		printf("%d", ctx->blkcount);
 #else
-		printf("%lld", ctx->blkcount);
+		printf("%lld", (long long)ctx->blkcount);
 #endif
 		break;
 	case 'c':
@@ -397,7 +419,7 @@ static _INLINE_ void expand_percent_expression(ext2_filsys fs, char ch,
 #ifdef EXT2_NO_64_TYPE
 		printf("%u", ctx->num);
 #else
-		printf("%llu", ctx->num);
+		printf("%llu", (long long)ctx->num);
 #endif
 		break;
 	case 'p':
@@ -414,7 +436,7 @@ static _INLINE_ void expand_percent_expression(ext2_filsys fs, char ch,
 		print_pathname(fs, ctx->dir, ctx->ino);
 		break;
 	case 'S':
-		printf("%d", get_backup_sb(NULL, fs, NULL, NULL));
+		printf("%u", get_backup_sb(NULL, fs, NULL, NULL));
 		break;
 	case 's':
 		printf("%s", ctx->str ? ctx->str : "NULL");
@@ -423,7 +445,7 @@ static _INLINE_ void expand_percent_expression(ext2_filsys fs, char ch,
 #ifdef EXT2_NO_64_TYPE
 		printf("0x%x", ctx->num);
 #else
-		printf("0x%llx", ctx->num);
+		printf("0x%llx", (long long)ctx->num);
 #endif
 		break;
 	default:
@@ -434,7 +456,8 @@ static _INLINE_ void expand_percent_expression(ext2_filsys fs, char ch,
 }	
 
 void print_e2fsck_message(e2fsck_t ctx, const char *msg,
-			  struct problem_context *pctx, int first)
+			  struct problem_context *pctx, int first,
+			  int recurse)
 {
 	ext2_filsys fs = ctx->fs;
 	const char *	cp;
@@ -444,7 +467,7 @@ void print_e2fsck_message(e2fsck_t ctx, const char *msg,
 	for (cp = msg; *cp; cp++) {
 		if (cp[0] == '@') {
 			cp++;
-			expand_at_expression(ctx, *cp, pctx, &first);
+			expand_at_expression(ctx, *cp, pctx, &first, recurse);
 		} else if (cp[0] == '%' && cp[1] == 'I') {
 			cp += 2;
 			expand_inode_expression(*cp, pctx);
