@@ -37,7 +37,7 @@
 #endif
 
 #ifndef HAVE_INTPTR_T
-typedef long intptr_t
+typedef long intptr_t;
 #endif
 
 /* Needed for architectures where sizeof(int) != sizeof(void *) */
@@ -96,6 +96,7 @@ static void pass1c(e2fsck_t ctx, char *block_buf);
 static void pass1d(e2fsck_t ctx, char *block_buf);
 
 static int dup_inode_count = 0;
+static int dup_inode_founddir = 0;
 
 static dict_t blk_dict, ino_dict;
 
@@ -146,7 +147,12 @@ static void add_dupe(e2fsck_t ctx, ext2_ino_t ino, blk_t blk,
 	else {
 		di = (struct dup_inode *) e2fsck_allocate_memory(ctx,
 			 sizeof(struct dup_inode), "duplicate inode header");
-		di->dir = (ino == EXT2_ROOT_INO) ? EXT2_ROOT_INO : 0 ;
+		if (ino == EXT2_ROOT_INO) {
+			di->dir = EXT2_ROOT_INO;
+			dup_inode_founddir++;
+		} else
+			di->dir = 0;
+
 		di->num_dupblocks = 0;
 		di->block_list = 0;
 		di->inode = *inode;
@@ -372,8 +378,10 @@ static int search_dirent_proc(ext2_ino_t dir, int entry,
 	if (!n)
 		return 0;
 	p = (struct dup_inode *) dnode_get(n);
-	p->dir = dir;
-	sd->count--;
+	if (!p->dir) {
+		p->dir = dir;
+		sd->count--;
+	}
 
 	return(sd->count ? 0 : DIRENT_ABORT);
 }
@@ -394,7 +402,7 @@ static void pass1c(e2fsck_t ctx, char *block_buf)
 	 * Search through all directories to translate inodes to names
 	 * (by searching for the containing directory for that inode.)
 	 */
-	sd.count = dup_inode_count;
+	sd.count = dup_inode_count - dup_inode_founddir;
 	sd.first_inode = EXT2_FIRST_INODE(fs->super);
 	sd.max_inode = fs->super->s_inodes_count;
 	ext2fs_dblist_dir_iterate(fs->dblist, 0, block_buf,
@@ -752,11 +760,26 @@ static int clone_file(e2fsck_t ctx, ext2_ino_t ino,
 		 * them to point to the new EA block.
 		 */
 		n = dict_lookup(&blk_dict, INT_TO_VOIDPTR(blk));
+		if (!n) {
+			com_err("clone_file", 0, 
+				_("internal error: couldn't lookup EA "
+				  "block record for %u"), blk);
+			retval = 0; /* OK to stumble on... */
+			goto errout;
+		}
 		db = (struct dup_block *) dnode_get(n);
 		for (ino_el = db->inode_list; ino_el; ino_el = ino_el->next) {
 			if (ino_el->inode == ino)
 				continue;
 			n = dict_lookup(&ino_dict, INT_TO_VOIDPTR(ino_el->inode));
+			if (!n) {
+				com_err("clone_file", 0, 
+					_("internal error: couldn't lookup EA "
+					  "inode record for %u"), 
+					ino_el->inode);
+				retval = 0; /* OK to stumble on... */
+				goto errout;
+			}
 			di = (struct dup_inode *) dnode_get(n);
 			if (di->inode.i_file_acl == blk) {
 				di->inode.i_file_acl = dp->inode.i_file_acl;
@@ -779,16 +802,16 @@ errout:
 static int check_if_fs_block(e2fsck_t ctx, blk_t test_block)
 {
 	ext2_filsys fs = ctx->fs;
-	blk_t	block;
+	blk_t	first_block;
 	dgrp_t	i;
 	
-	block = fs->super->s_first_data_block;
+	first_block = fs->super->s_first_data_block;
 	for (i = 0; i < fs->group_desc_count; i++) {
 
-		/* Check superblocks/block group descriptros */
+		/* Check superblocks/block group descriptors */
 		if (ext2fs_bg_has_super(fs, i)) {
-			if (test_block >= block &&
-			    (test_block <= block + fs->desc_blocks))
+			if (test_block >= first_block &&
+			    (test_block <= first_block + fs->desc_blocks))
 				return 1;
 		}
 		
@@ -804,7 +827,7 @@ static int check_if_fs_block(e2fsck_t ctx, blk_t test_block)
 		    (test_block == fs->group_desc[i].bg_inode_bitmap))
 			return 1;
 		
-		block += fs->super->s_blocks_per_group;
+		first_block += fs->super->s_blocks_per_group;
 	}
 	return 0;
 }
