@@ -79,15 +79,16 @@ struct out_dir {
 static int fill_dir_block(ext2_filsys fs,
 			  blk_t	*block_nr,
 			  e2_blkcnt_t blockcnt,
-			  blk_t ref_block,
-			  int ref_offset, 
+			  blk_t ref_block EXT2FS_ATTR((unused)),
+			  int ref_offset EXT2FS_ATTR((unused)),
 			  void *priv_data)
 {
 	struct fill_dir_struct	*fd = (struct fill_dir_struct *) priv_data;
 	struct hash_entry 	*new_array, *ent;
 	struct ext2_dir_entry 	*dirent;
 	char			*dir;
-	int			offset, dir_offset;
+	unsigned int		offset, dir_offset;
+	int			hash_alg;
 	
 	if (blockcnt < 0)
 		return 0;
@@ -107,6 +108,10 @@ static int fill_dir_block(ext2_filsys fs,
 		if (fd->err)
 			return BLOCK_ABORT;
 	}
+	hash_alg = fs->super->s_def_hash_version;
+	if ((hash_alg <= EXT2_HASH_TEA) &&
+	    (fs->super->s_flags & EXT2_FLAGS_UNSIGNED_HASH))
+		hash_alg += 3;
 	/* While the directory block is "hot", index it. */
 	dir_offset = 0;
 	while (dir_offset < fs->blocksize) {
@@ -145,8 +150,7 @@ static int fill_dir_block(ext2_filsys fs,
 		if (fd->compress)
 			ent->hash = ent->minor_hash = 0;
 		else {
-			fd->err = ext2fs_dirhash(fs->super->s_def_hash_version,
-						 dirent->name,
+			fd->err = ext2fs_dirhash(hash_alg, dirent->name,
 						 dirent->name_len & 0xFF,
 						 fs->super->s_hash_seed,
 						 &ent->hash, &ent->minor_hash);
@@ -323,10 +327,16 @@ static int duplicate_search_and_fix(e2fsck_t ctx, ext2_filsys fs,
 	int			fixed = 0;
 	char			new_name[256];
 	__u16			new_len;
+	int			hash_alg;
 	
 	clear_problem_context(&pctx);
 	pctx.ino = ino;
 
+	hash_alg = fs->super->s_def_hash_version;
+	if ((hash_alg <= EXT2_HASH_TEA) &&
+	    (fs->super->s_flags & EXT2_FLAGS_UNSIGNED_HASH))
+		hash_alg += 3;
+
 	for (i=1; i < fd->num_array; i++) {
 		ent = fd->harray + i;
 		prev = ent - 1;
@@ -363,8 +373,7 @@ static int duplicate_search_and_fix(e2fsck_t ctx, ext2_filsys fs,
 		if (fix_problem(ctx, PR_2_NON_UNIQUE_FILE, &pctx)) {
 			memcpy(ent->dir->name, new_name, new_len & 0xFF);
 			ent->dir->name_len = new_len;
-			ext2fs_dirhash(fs->super->s_def_hash_version,
-				       ent->dir->name,
+			ext2fs_dirhash(hash_alg, ent->dir->name,
 				       ent->dir->name_len & 0xFF,
 				       fs->super->s_hash_seed,
 				       &ent->hash, &ent->minor_hash);
@@ -589,8 +598,8 @@ struct write_dir_struct {
 static int write_dir_block(ext2_filsys fs,
 			   blk_t	*block_nr,
 			   e2_blkcnt_t blockcnt,
-			   blk_t ref_block,
-			   int ref_offset, 
+			   blk_t ref_block EXT2FS_ATTR((unused)),
+			   int ref_offset EXT2FS_ATTR((unused)), 
 			   void *priv_data)
 {
 	struct write_dir_struct	*wd = (struct write_dir_struct *) priv_data;
@@ -763,9 +772,10 @@ void e2fsck_rehash_directories(e2fsck_t ctx)
 #endif
 	struct dir_info		*dir;
 	ext2_u32_iterate 	iter;
+	struct dir_info_iter *	dirinfo_iter = 0;
 	ext2_ino_t		ino;
 	errcode_t		retval;
-	int			i, cur, max, all_dirs, dir_index, first = 1;
+	int			cur, max, all_dirs, dir_index, first = 1;
 
 #ifdef RESOURCE_TRACK
 	init_resource_track(&rtrack);
@@ -783,7 +793,7 @@ void e2fsck_rehash_directories(e2fsck_t ctx)
 	dir_index = ctx->fs->super->s_feature_compat & EXT2_FEATURE_COMPAT_DIR_INDEX;
 	cur = 0;
 	if (all_dirs) {
-		i = 0;
+		dirinfo_iter = e2fsck_dir_info_iter_begin(ctx);
 		max = e2fsck_get_num_dirinfo(ctx);
 	} else {
 		retval = ext2fs_u32_list_iterate_begin(ctx->dirs_to_hash, 
@@ -797,7 +807,8 @@ void e2fsck_rehash_directories(e2fsck_t ctx)
 	}
 	while (1) {
 		if (all_dirs) {
-			if ((dir = e2fsck_dir_info_iter(ctx, &i)) == 0)
+			if ((dir = e2fsck_dir_info_iter(ctx, 
+							dirinfo_iter)) == 0)
 				break;
 			ino = dir->ino;
 		} else {
@@ -824,7 +835,9 @@ void e2fsck_rehash_directories(e2fsck_t ctx)
 			       100.0 * (float) (++cur) / (float) max, ino);
 	}
 	end_problem_latch(ctx, PR_LATCH_OPTIMIZE_DIR);
-	if (!all_dirs)
+	if (all_dirs)
+		e2fsck_dir_info_iter_end(ctx, dirinfo_iter);
+	else
 		ext2fs_u32_list_iterate_end(iter);
 	
 	if (ctx->dirs_to_hash)
