@@ -38,19 +38,6 @@ static void check_super_value(e2fsck_t ctx, const char *descr,
 	}
 }
 
-/*
- * This routine may get stubbed out in special compilations of the
- * e2fsck code..
- */
-#ifndef EXT2_SPECIAL_DEVICE_SIZE
-errcode_t e2fsck_get_device_size(e2fsck_t ctx)
-{
-	return (ext2fs_get_device_size(ctx->filesystem_name,
-				       EXT2_BLOCK_SIZE(ctx->fs->super),
-				       &ctx->num_blocks));
-}
-#endif
-
 /*
  * helper function to release an inode
  */
@@ -220,7 +207,7 @@ static int release_inode_blocks(e2fsck_t ctx, ext2_ino_t ino,
 		}
 		if (retval) {
 			com_err("release_inode_blocks", retval,
-		_("while calling ext2fs_adjust_ea_refocunt for inode %d"),
+		_("while calling ext2fs_adjust_ea_refcount for inode %d"),
 				ino);
 			return 1;
 		}
@@ -318,12 +305,13 @@ return_abort:
  * resize inode should be cleared) as well as the case where on-line
  * resizing is enabled.
  */
-void check_resize_inode(e2fsck_t ctx)
+static void check_resize_inode(e2fsck_t ctx)
 {
 	ext2_filsys fs = ctx->fs;
 	struct ext2_inode inode;
 	struct problem_context	pctx;
-	int		i, j, gdt_off, ind_off;
+	int		i, gdt_off, ind_off;
+	dgrp_t		j;
 	blk_t		blk, pblk, expect;
 	__u32 		*dind_buf = 0, *ind_buf;
 	errcode_t	retval;
@@ -436,6 +424,34 @@ cleanup:
 
  }
 
+/*
+ * This function checks the dirhash signed/unsigned hint if necessary.
+ */
+static void e2fsck_fix_dirhash_hint(e2fsck_t ctx)
+{
+	struct ext2_super_block *sb = ctx->fs->super;
+	struct problem_context pctx;
+	char	c;
+
+	if ((ctx->options & E2F_OPT_READONLY) ||
+	    !(sb->s_feature_compat & EXT2_FEATURE_COMPAT_DIR_INDEX) ||
+	    (sb->s_flags & (EXT2_FLAGS_SIGNED_HASH|EXT2_FLAGS_UNSIGNED_HASH)))
+		return;
+
+	c = (char) 255;
+
+	clear_problem_context(&pctx);
+	if (fix_problem(ctx, PR_0_DIRHASH_HINT, &pctx)) {
+		if (((int) c) == -1) {
+			sb->s_flags |= EXT2_FLAGS_SIGNED_HASH;
+		} else {
+			sb->s_flags |= EXT2_FLAGS_UNSIGNED_HASH;
+		}
+		ext2fs_mark_super_dirty(ctx->fs);
+	}
+}
+
+
 void check_super_block(e2fsck_t ctx)
 {
 	ext2_filsys fs = ctx->fs;
@@ -507,22 +523,14 @@ void check_super_block(e2fsck_t ctx)
 		return;
 	}
 		
-	if (!ctx->num_blocks) {
-		pctx.errcode = e2fsck_get_device_size(ctx);
-		if (pctx.errcode && pctx.errcode != EXT2_ET_UNIMPLEMENTED) {
-			fix_problem(ctx, PR_0_GETSIZE_ERROR, &pctx);
+	if ((ctx->flags & E2F_FLAG_GOT_DEVSIZE) &&
+	    (ctx->num_blocks < sb->s_blocks_count)) {
+		pctx.blk = sb->s_blocks_count;
+		pctx.blk2 = ctx->num_blocks;
+		if (fix_problem(ctx, PR_0_FS_SIZE_WRONG, &pctx)) {
 			ctx->flags |= E2F_FLAG_ABORT;
 			return;
 		}
-		if ((pctx.errcode != EXT2_ET_UNIMPLEMENTED) &&
-		    (ctx->num_blocks < sb->s_blocks_count)) {
-			pctx.blk = sb->s_blocks_count;
-			pctx.blk2 = ctx->num_blocks;
-			if (fix_problem(ctx, PR_0_FS_SIZE_WRONG, &pctx)) {
-				ctx->flags |= E2F_FLAG_ABORT;
-				return;
-			}
-		}
 	}
 
 	if (sb->s_log_block_size != (__u32) sb->s_log_frag_size) {
@@ -566,15 +574,15 @@ void check_super_block(e2fsck_t ctx)
 	 * Verify the group descriptors....
 	 */
 	first_block =  sb->s_first_data_block;
-	last_block = first_block + blocks_per_group;
 
 	for (i = 0, gd=fs->group_desc; i < fs->group_desc_count; i++, gd++) {
 		pctx.group = i;
-		
-		if (i == fs->group_desc_count - 1)
-			last_block = sb->s_blocks_count;
+
+		first_block = ext2fs_group_first_block(fs, i);
+		last_block = ext2fs_group_last_block(fs, i);
+
 		if ((gd->bg_block_bitmap < first_block) ||
-		    (gd->bg_block_bitmap >= last_block)) {
+		    (gd->bg_block_bitmap > last_block)) {
 			pctx.blk = gd->bg_block_bitmap;
 			if (fix_problem(ctx, PR_0_BB_NOT_GROUP, &pctx))
 				gd->bg_block_bitmap = 0;
@@ -584,7 +592,7 @@ void check_super_block(e2fsck_t ctx)
 			ctx->invalid_bitmaps++;
 		}
 		if ((gd->bg_inode_bitmap < first_block) ||
-		    (gd->bg_inode_bitmap >= last_block)) {
+		    (gd->bg_inode_bitmap > last_block)) {
 			pctx.blk = gd->bg_inode_bitmap;
 			if (fix_problem(ctx, PR_0_IB_NOT_GROUP, &pctx))
 				gd->bg_inode_bitmap = 0;
@@ -595,7 +603,7 @@ void check_super_block(e2fsck_t ctx)
 		}
 		if ((gd->bg_inode_table < first_block) ||
 		    ((gd->bg_inode_table +
-		      fs->inode_blocks_per_group - 1) >= last_block)) {
+		      fs->inode_blocks_per_group - 1) > last_block)) {
 			pctx.blk = gd->bg_inode_table;
 			if (fix_problem(ctx, PR_0_ITABLE_NOT_GROUP, &pctx))
 				gd->bg_inode_table = 0;
@@ -606,8 +614,6 @@ void check_super_block(e2fsck_t ctx)
 		}
 		free_blocks += gd->bg_free_blocks_count;
 		free_inodes += gd->bg_free_inodes_count;
-		first_block += sb->s_blocks_per_group;
-		last_block += sb->s_blocks_per_group;
 
 		if ((gd->bg_free_blocks_count > sb->s_blocks_per_group) ||
 		    (gd->bg_free_inodes_count > sb->s_inodes_per_group) ||
@@ -709,14 +715,14 @@ void check_super_block(e2fsck_t ctx)
 	 * Check to see if the superblock last mount time or last
 	 * write time is in the future.
 	 */
-	if (fs->super->s_mtime > ctx->now) {
+	if (fs->super->s_mtime > (__u32) ctx->now) {
 		pctx.num = fs->super->s_mtime;
 		if (fix_problem(ctx, PR_0_FUTURE_SB_LAST_MOUNT, &pctx)) {
 			fs->super->s_mtime = ctx->now;
 			ext2fs_mark_super_dirty(fs);
 		}
 	}
-	if (fs->super->s_wtime > ctx->now) {
+	if (fs->super->s_wtime > (__u32) ctx->now) {
 		pctx.num = fs->super->s_wtime;
 		if (fix_problem(ctx, PR_0_FUTURE_SB_LAST_WRITE, &pctx)) {
 			fs->super->s_wtime = ctx->now;
@@ -734,5 +740,10 @@ void check_super_block(e2fsck_t ctx)
 	 */
 	e2fsck_fix_ext3_journal_hint(ctx);
 
+	/*
+	 * Add dirhash hint if necessary
+	 */
+	e2fsck_fix_dirhash_hint(ctx);
+
 	return;
 }
