@@ -46,7 +46,7 @@ static int figure_label_len(const unsigned char *label, int len)
 }
 
 static unsigned char *get_buffer(struct blkid_probe *pr, 
-			  unsigned off, size_t len)
+			  blkid_loff_t off, size_t len)
 {
 	ssize_t		ret_read;
 	unsigned char	*newbuf;
@@ -74,7 +74,7 @@ static unsigned char *get_buffer(struct blkid_probe *pr,
 			pr->buf = newbuf;
 			pr->buf_max = len;
 		}
-		if (lseek(pr->fd, off, SEEK_SET) < 0)
+		if (blkid_llseek(pr->fd, off, SEEK_SET) < 0)
 			return NULL;
 		ret_read = read(pr->fd, pr->buf, len);
 		if (ret_read != (ssize_t) len)
@@ -105,7 +105,7 @@ static int check_mdraid(int fd, unsigned char *ret_uuid)
 		return -BLKID_ERR_IO;
 
 	/* Check for magic number */
-	if (memcmp("\251+N\374", buf, 4))
+	if (memcmp("\251+N\374", buf, 4) && memcmp("\374N+\251", buf, 4))
 		return -BLKID_ERR_PARAM;
 
 	if (!ret_uuid)
@@ -116,7 +116,7 @@ static int check_mdraid(int fd, unsigned char *ret_uuid)
 	md = (struct mdp_superblock_s *)buf;
 	if (md->set_uuid0 || md->set_uuid1 || md->set_uuid2 || md->set_uuid3) {
 		memcpy(ret_uuid, &md->set_uuid0, 4);
-		memcpy(ret_uuid, &md->set_uuid1, 12);
+		memcpy(ret_uuid + 4, &md->set_uuid1, 12);
 	}
 	return 0;
 }
@@ -153,7 +153,6 @@ static int probe_ext3(struct blkid_probe *probe,
 		      unsigned char *buf)
 {
 	struct ext2_super_block *es;
-	char uuid[37];
 	es = (struct ext2_super_block *)buf;
 
 	/* Distinguish between jbd and ext2/3 fs */
@@ -225,7 +224,7 @@ static char *no_name = "NO NAME    ";
 
 static unsigned char *search_fat_label(struct vfat_dir_entry *dir, int count)
 {
-	unsigned int i;
+	int i;
 
 	for (i = 0; i < count; i++) {
 		if (dir[i].name[0] == 0x00)
@@ -254,7 +253,7 @@ static int probe_fat(struct blkid_probe *probe,
 	struct msdos_super_block *ms = (struct msdos_super_block *) buf;
 	struct vfat_dir_entry *dir;
 	char serno[10];
-	const unsigned char *label = 0, *vol_label = 0;
+	const unsigned char *label = 0, *vol_label = 0, *tmp;
 	unsigned char	*vol_serno;
 	int label_len = 0, maxloop = 100;
 	__u16 sector_size, dir_entries, reserved;
@@ -262,14 +261,17 @@ static int probe_fat(struct blkid_probe *probe,
 	__u32 buf_size, start_data_sect, next, root_start, root_dir_entries;
 
 	/* sector size check */
-	sector_size = blkid_le16(*((__u16 *) &ms->ms_sector_size));
+	tmp = (unsigned char *)&ms->ms_sector_size;
+	sector_size = tmp[0] + (tmp[1] << 8);
 	if (sector_size != 0x200 && sector_size != 0x400 &&
 	    sector_size != 0x800 && sector_size != 0x1000)
 		return 1;
 
-	dir_entries = blkid_le16(*((__u16 *) &ms->ms_dir_entries));
+	tmp = (unsigned char *)&ms->ms_dir_entries;
+	dir_entries = tmp[0] + (tmp[1] << 8);
 	reserved =  blkid_le16(ms->ms_reserved);
-	sect_count = blkid_le16(*((__u16 *) &ms->ms_sectors));
+	tmp = (unsigned char *)&ms->ms_sectors;
+	sect_count = tmp[0] + (tmp[1] << 8);
 	if (sect_count == 0)
 		sect_count = blkid_le32(ms->ms_total_sect);
 
@@ -348,8 +350,8 @@ static int probe_fat(struct blkid_probe *probe,
 	}
 
 	if (vol_label && memcmp(vol_label, no_name, 11)) {
-		label = vol_label;
-		label_len = figure_label_len(vol_label, 11);
+		if ((label_len = figure_label_len(vol_label, 11)))
+			label = vol_label;
 	}
 
 	/* We can't just print them as %04X, because they are unaligned */
@@ -396,6 +398,114 @@ static int probe_fat_nomagic(struct blkid_probe *probe,
 	return probe_fat(probe, id, buf);
 }
 
+static int probe_ntfs(struct blkid_probe *probe,
+		      struct blkid_magic *id __BLKID_ATTR((unused)), 
+		      unsigned char *buf)
+{
+	struct ntfs_super_block *ns;
+	struct master_file_table_record *mft;
+	struct file_attribute *attr;
+	char		uuid_str[17], label_str[129], *cp;
+	int		bytes_per_sector, sectors_per_cluster;
+	int		mft_record_size, attr_off, attr_len;
+	unsigned int	i, attr_type, val_len;
+	int		val_off;
+	__u64		nr_clusters;
+	blkid_loff_t off;
+	unsigned char *buf_mft, *val;
+
+	ns = (struct ntfs_super_block *) buf;
+
+	bytes_per_sector = ns->bios_parameter_block[0] +
+		(ns->bios_parameter_block[1]  << 8);
+	sectors_per_cluster = ns->bios_parameter_block[2];
+
+	if ((bytes_per_sector < 512) || (sectors_per_cluster == 0))
+		return 1;
+
+	if (ns->cluster_per_mft_record < 0)
+		mft_record_size = 1 << (0-ns->cluster_per_mft_record);
+	else
+		mft_record_size = ns->cluster_per_mft_record * 
+			sectors_per_cluster * bytes_per_sector;
+	nr_clusters = blkid_le64(ns->number_of_sectors) / sectors_per_cluster;
+
+	if ((blkid_le64(ns->mft_cluster_location) > nr_clusters) ||
+	    (blkid_le64(ns->mft_mirror_cluster_location) > nr_clusters))
+		return 1;
+
+	off = blkid_le64(ns->mft_mirror_cluster_location) * 
+		bytes_per_sector * sectors_per_cluster;
+
+	buf_mft = get_buffer(probe, off, mft_record_size);
+	if (!buf_mft)
+		return 1;
+
+	if (memcmp(buf_mft, "FILE", 4))
+		return 1;
+
+	off = blkid_le64(ns->mft_cluster_location) * bytes_per_sector * 
+		sectors_per_cluster;
+
+	buf_mft = get_buffer(probe, off, mft_record_size);
+	if (!buf_mft)
+		return 1;
+
+	if (memcmp(buf_mft, "FILE", 4))
+		return 1;
+
+	off += MFT_RECORD_VOLUME * mft_record_size;
+
+	buf_mft = get_buffer(probe, off, mft_record_size);
+	if (!buf_mft)
+		return 1;
+
+	if (memcmp(buf_mft, "FILE", 4))
+		return 1;
+
+	mft = (struct master_file_table_record *) buf_mft;
+
+	attr_off = blkid_le16(mft->attrs_offset);
+	label_str[0] = 0;
+	
+	while (1) {
+		attr = (struct file_attribute *) (buf_mft + attr_off);
+		attr_len = blkid_le16(attr->len);
+		attr_type = blkid_le32(attr->type);
+		val_off = blkid_le16(attr->value_offset);
+		val_len = blkid_le32(attr->value_len);
+
+		attr_off += attr_len;
+
+		if ((attr_off > mft_record_size) ||
+		    (attr_len == 0))
+			break;
+
+		if (attr_type == MFT_RECORD_ATTR_END)
+			break;
+
+		if (attr_type == MFT_RECORD_ATTR_VOLUME_NAME) {
+			if (val_len > sizeof(label_str))
+				val_len = sizeof(label_str)-1;
+
+			for (i=0, cp=label_str; i < val_len; i+=2,cp++) {
+				val = ((__u8 *) attr) + val_off + i;
+				*cp = val[0];
+				if (val[1])
+					*cp = '?';
+			}
+			*cp = 0;
+		}
+	}
+
+	sprintf(uuid_str, "%llX", blkid_le64(ns->volume_serial));
+	blkid_set_tag(probe->dev, "UUID", uuid_str, 0);
+	if (label_str[0])
+		blkid_set_tag(probe->dev, "LABEL", label_str, 0);
+	return 0;
+}
+
+
 static int probe_xfs(struct blkid_probe *probe,
 		     struct blkid_magic *id __BLKID_ATTR((unused)), 
 		     unsigned char *buf)
@@ -468,6 +578,18 @@ static int probe_jfs(struct blkid_probe *probe,
 	return 0;
 }
 
+static int probe_luks(struct blkid_probe *probe,
+		       struct blkid_magic *id __BLKID_ATTR((unused)),
+		       unsigned char *buf)
+{
+	unsigned char uuid[40];
+	/* 168 is the offset to the 40 character uuid:
+	 * http://luks.endorphin.org/LUKS-on-disk-format.pdf */
+	strncpy(uuid, buf+168, 40);
+	blkid_set_tag(probe->dev, "UUID", uuid, sizeof(uuid));
+	return 0;
+}
+
 static int probe_romfs(struct blkid_probe *probe,
 		       struct blkid_magic *id __BLKID_ATTR((unused)), 
 		       unsigned char *buf)
@@ -587,10 +709,10 @@ static int probe_udf(struct blkid_probe *probe,
 		   NSR01 for UDF 1.00
 		   NSR02 for UDF 1.50
 		   NSR03 for UDF 2.00 */
-		if (!strncmp(isosb->vd_id, "NSR0", 4))
+		if (!memcmp(isosb->vd_id, "NSR0", 4))
 			return 0;
 		for (m = udf_magic; *m; m++)
-			if (!strncmp(*m, isosb->vd_id, 5))
+			if (!memcmp(*m, isosb->vd_id, 5))
 				break;
 		if (*m == 0)
 			return 1;
@@ -646,6 +768,50 @@ static int probe_oracleasm(struct blkid_probe *probe,
 	return 0;
 }
 
+static int probe_gfs(struct blkid_probe *probe,
+		     struct blkid_magic *id __BLKID_ATTR((unused)),
+		     unsigned char *buf)
+{
+	struct gfs2_sb *sbd;
+	const char *label = 0;
+
+	sbd = (struct gfs2_sb *)buf;
+
+	if (blkid_be32(sbd->sb_fs_format) == GFS_FORMAT_FS &&
+	    blkid_be32(sbd->sb_multihost_format) == GFS_FORMAT_MULTI)
+	{	
+		blkid_set_tag(probe->dev, "UUID", 0, 0);
+	
+		if (strlen(sbd->sb_locktable))
+			label = sbd->sb_locktable;
+		blkid_set_tag(probe->dev, "LABEL", label, sizeof(sbd->sb_locktable));
+		return 0;
+	}
+	return 1;
+}
+
+static int probe_gfs2(struct blkid_probe *probe,
+		     struct blkid_magic *id __BLKID_ATTR((unused)),
+		     unsigned char *buf)
+{
+	struct gfs2_sb *sbd;
+	const char *label = 0;
+
+	sbd = (struct gfs2_sb *)buf;
+
+	if (blkid_be32(sbd->sb_fs_format) == GFS2_FORMAT_FS &&
+	    blkid_be32(sbd->sb_multihost_format) == GFS2_FORMAT_MULTI)
+	{	
+		blkid_set_tag(probe->dev, "UUID", 0, 0);
+	
+		if (strlen(sbd->sb_locktable))
+			label = sbd->sb_locktable;
+		blkid_set_tag(probe->dev, "LABEL", label, sizeof(sbd->sb_locktable));
+		return 0;
+	}
+	return 1;
+}
+
 /*
  * BLKID_BLK_OFFS is at least as large as the highest bim_kboff defined
  * in the type_array table below + bim_kbalign.
@@ -663,7 +829,7 @@ static int probe_oracleasm(struct blkid_probe *probe,
 static struct blkid_magic type_array[] = {
 /*  type     kboff   sboff len  magic			probe */
   { "oracleasm", 0,	32,  8, "ORCLDISK",		probe_oracleasm },
-  { "ntfs",      0,      3,  8, "NTFS    ",             0 },
+  { "ntfs",	 0,	 3,  8, "NTFS    ",		probe_ntfs },
   { "jbd",	 1,   0x38,  2, "\123\357",		probe_jbd },
   { "ext3",	 1,   0x38,  2, "\123\357",		probe_ext3 },
   { "ext2",	 1,   0x38,  2, "\123\357",		probe_ext2 },
@@ -673,6 +839,8 @@ static struct blkid_magic type_array[] = {
   { "reiserfs", 64,   0x34,  8, "ReIsErFs",		probe_reiserfs },
   { "reiserfs",	 8,	20,  8, "ReIsErFs",		probe_reiserfs },
   { "reiser4",  64,	 0,  7, "ReIsEr4",		probe_reiserfs4 },
+  { "gfs2",     64,      0,  4, "\x01\x16\x19\x70",     probe_gfs2 },
+  { "gfs",      64,      0,  4, "\x01\x16\x19\x70",     probe_gfs },
   { "vfat",      0,   0x52,  5, "MSWIN",                probe_fat },
   { "vfat",      0,   0x52,  8, "FAT32   ",             probe_fat },
   { "vfat",      0,   0x36,  5, "MSDOS",                probe_fat },
@@ -729,6 +897,7 @@ static struct blkid_magic type_array[] = {
   { "ocfs2",	 2,	 0,  6,	"OCFSV2",		probe_ocfs2 },
   { "ocfs2",	 4,	 0,  6,	"OCFSV2",		probe_ocfs2 },
   { "ocfs2",	 8,	 0,  6,	"OCFSV2",		probe_ocfs2 },
+  { "crypt_LUKS", 0,	 0,  6,	"LUKS\xba\xbe",		probe_luks },
   {   NULL,	 0,	 0,  0, NULL,			NULL }
 };
 
@@ -765,8 +934,8 @@ blkid_dev blkid_verify(blkid_cache cache, blkid_dev dev)
 		return dev;
 
 	DBG(DEBUG_PROBE,
-	    printf("need to revalidate %s (time since last check %lu)\n", 
-		   dev->bid_name, diff));
+	    printf("need to revalidate %s (time since last check %llu)\n", 
+		   dev->bid_name, (unsigned long long)diff));
 
 	if (((probe.fd = open(dev->bid_name, O_RDONLY)) < 0) ||
 	    (fstat(probe.fd, &st) < 0)) {
@@ -830,6 +999,9 @@ try_again:
 		/*
 		 * Zap the device filesystem information and try again
 		 */
+		DBG(DEBUG_PROBE,
+		    printf("previous fs type %s not valid, "
+			   "trying full probe\n", dev->bid_type));
 		iter = blkid_tag_iterate_begin(dev);
 		while (blkid_tag_next(iter, &type, &value) == 0)
 			blkid_set_tag(dev, type, 0, 0);
@@ -838,9 +1010,9 @@ try_again:
 	}
 
 	if (!dev->bid_type) {
-		if (probe.fd >= 0) close(probe.fd);
 		blkid_free_dev(dev);
-		return NULL;
+		dev = 0;
+		goto found_type;
 	}
 		
 found_type:
@@ -853,14 +1025,15 @@ found_type:
 		blkid_set_tag(dev, "TYPE", type, 0);
 				
 		DBG(DEBUG_PROBE, printf("%s: devno 0x%04llx, type %s\n",
-			   dev->bid_name, st.st_rdev, type));
+			   dev->bid_name, (long long)st.st_rdev, type));
 	}
 
 	if (probe.sbbuf)
 		free(probe.sbbuf);
 	if (probe.buf)
 		free(probe.buf);
-	close(probe.fd);
+	if (probe.fd >= 0) 
+		close(probe.fd);
 
 	return dev;
 }
