@@ -23,23 +23,38 @@
 #include <errno.h>
 #endif
 #include <fcntl.h>
-#ifdef HAVE_LINUX_FD_H
+#ifdef HAVE_SYS_IOCTL_H
 #include <sys/ioctl.h>
+#endif
+#ifdef HAVE_LINUX_FD_H
 #include <linux/fd.h>
 #endif
 #ifdef HAVE_SYS_DISKLABEL_H
-#include <sys/ioctl.h>
 #include <sys/disklabel.h>
-#endif /* HAVE_SYS_DISKLABEL_H */
+#endif
+#ifdef HAVE_SYS_DISK_H
+#ifdef HAVE_SYS_QUEUE_H
+#include <sys/queue.h> /* for LIST_HEAD */
+#endif
+#include <sys/disk.h>
+#endif
+#ifdef __linux__
+#include <sys/utsname.h>
+#endif
+#if HAVE_SYS_STAT_H
+#include <sys/stat.h>
+#endif
+#include <ctype.h>
 
 #if defined(__linux__) && defined(_IO) && !defined(BLKGETSIZE)
 #define BLKGETSIZE _IO(0x12,96)	/* return device size */
 #endif
 
-#ifdef APPLE_DARWIN
-#include <sys/ioctl.h>
-#include <sys/disk.h>
+#if defined(__linux__) && defined(_IOR) && !defined(BLKGETSIZE64)
+#define BLKGETSIZE64 _IOR(0x12,114,size_t)	/* return device size in bytes (u64 *arg) */
+#endif
 
+#ifdef APPLE_DARWIN
 #define BLKGETSIZE DKIOCGETBLOCKCOUNT32
 #endif /* APPLE_DARWIN */
 
@@ -50,6 +65,10 @@
 #include "windows.h"
 #include "winioctl.h"
 
+#if (_WIN32_WINNT >= 0x0500)
+#define HAVE_GET_FILE_SIZE_EX 1
+#endif
+
 errcode_t ext2fs_get_device_size(const char *file, int blocksize,
 				 blk_t *retblocks)
 {
@@ -57,7 +76,11 @@ errcode_t ext2fs_get_device_size(const char *file, int blocksize,
 	PARTITION_INFORMATION pi;
 	DISK_GEOMETRY gi;
 	DWORD retbytes;
+#ifdef HAVE_GET_FILE_SIZE_EX
 	LARGE_INTEGER filesize;
+#else
+	DWORD filesize;
+#endif /* HAVE_GET_FILE_SIZE_EX */
 
 	dev = CreateFile(file, GENERIC_READ, 
 			 FILE_SHARE_READ | FILE_SHARE_WRITE ,
@@ -82,9 +105,18 @@ errcode_t ext2fs_get_device_size(const char *file, int blocksize,
 			     gi.TracksPerCylinder *
 			     gi.Cylinders.QuadPart / blocksize;
 
+#ifdef HAVE_GET_FILE_SIZE_EX
 	} else if (GetFileSizeEx(dev, &filesize)) {
 		*retblocks = filesize.QuadPart / blocksize;
 	}
+#else
+	} else {
+		filesize = GetFileSize(dev, NULL);
+		if (INVALID_FILE_SIZE != filesize) {
+			*retblocks = filesize / blocksize;
+		}
+	}
+#endif /* HAVE_GET_FILE_SIZE_EX */
 
 	CloseHandle(dev);
 	return 0;
@@ -109,10 +141,13 @@ static int valid_offset (int fd, ext2_loff_t offset)
 errcode_t ext2fs_get_device_size(const char *file, int blocksize,
 				 blk_t *retblocks)
 {
-	int	fd;
-#ifdef BLKGETSIZE
-	unsigned long	size;
+	int	fd, rc = 0;
+	int valid_blkgetsize64 = 1;
+#ifdef __linux__
+	struct 		utsname ut;
 #endif
+	unsigned long long size64;
+	unsigned long	size;
 	ext2_loff_t high, low;
 #ifdef FDGETPRM
 	struct floppy_struct this_floppy;
@@ -132,21 +167,63 @@ errcode_t ext2fs_get_device_size(const char *file, int blocksize,
 	if (fd < 0)
 		return errno;
 
+#ifdef DKIOCGETBLOCKCOUNT	/* For Apple Darwin */
+	if (ioctl(fd, DKIOCGETBLOCKCOUNT, &size64) >= 0) {
+		if ((sizeof(*retblocks) < sizeof(unsigned long long))
+		    && ((size64 / (blocksize / 512)) > 0xFFFFFFFF)) {
+			rc = EFBIG;
+			goto out;
+		}
+		*retblocks = size64 / (blocksize / 512);
+		goto out;
+	}
+#endif
+
+#ifdef BLKGETSIZE64
+#ifdef __linux__
+	if ((uname(&ut) == 0) &&
+	    ((ut.release[0] == '2') && (ut.release[1] == '.') &&
+	     (ut.release[2] < '6') && (ut.release[3] == '.')))
+		valid_blkgetsize64 = 0;
+#endif
+	if (valid_blkgetsize64 &&
+	    ioctl(fd, BLKGETSIZE64, &size64) >= 0) {
+		if ((sizeof(*retblocks) < sizeof(unsigned long long)) &&
+		    ((size64 / blocksize) > 0xFFFFFFFF)) {
+			rc = EFBIG;
+			goto out;
+		}
+		*retblocks = size64 / blocksize;
+		goto out;
+	}
+#endif
+
 #ifdef BLKGETSIZE
 	if (ioctl(fd, BLKGETSIZE, &size) >= 0) {
-		close(fd);
 		*retblocks = size / (blocksize / 512);
-		return 0;
+		goto out;
 	}
 #endif
+
 #ifdef FDGETPRM
 	if (ioctl(fd, FDGETPRM, &this_floppy) >= 0) {
-		close(fd);
 		*retblocks = this_floppy.size / (blocksize / 512);
-		return 0;
+		goto out;
 	}
 #endif
+
 #ifdef HAVE_SYS_DISKLABEL_H
+#if defined(DIOCGMEDIASIZE)
+	{
+	    off_t ms;
+	    u_int bs;
+	    if (ioctl(fd, DIOCGMEDIASIZE, &ms) >= 0) {
+		*retblocks = ms / blocksize;
+		goto out;
+	    }
+	}
+#elif defined(DIOCGDINFO)
+	/* old disklabel interface */
 	part = strlen(file) - 1;
 	if (part >= 0) {
 		ch = file[part];
@@ -160,13 +237,32 @@ errcode_t ext2fs_get_device_size(const char *file, int blocksize,
 	if (part >= 0 && (ioctl(fd, DIOCGDINFO, (char *)&lab) >= 0)) {
 		pp = &lab.d_partitions[part];
 		if (pp->p_size) {
-			close(fd);
 			*retblocks = pp->p_size / (blocksize / 512);
-			return 0;
+			goto out;
 		}
 	}
+#endif /* defined(DIOCG*) */
 #endif /* HAVE_SYS_DISKLABEL_H */
 
+	{
+#ifdef HAVE_FSTAT64
+		struct stat64   st;
+		if (fstat64(fd, &st) == 0)
+#else
+		struct stat	st;
+		if (fstat(fd, &st) == 0)
+#endif
+			if (S_ISREG(st.st_mode)) {
+				if ((sizeof(*retblocks) < sizeof(unsigned long long)) &&
+				    ((st.st_size / blocksize) > 0xFFFFFFFF)) {
+					rc = EFBIG;
+					goto out;
+				}
+				*retblocks = st.st_size / blocksize;
+				goto out;
+			}
+	}
+
 	/*
 	 * OK, we couldn't figure it out by using a specialized ioctl,
 	 * which is generally the best way.  So do binary search to
@@ -185,9 +281,16 @@ errcode_t ext2fs_get_device_size(const char *file, int blocksize,
 			high = mid;
 	}
 	valid_offset (fd, 0);
+	size64 = low + 1;
+	if ((sizeof(*retblocks) < sizeof(unsigned long long))
+	    && ((size64 / blocksize) > 0xFFFFFFFF)) {
+		rc = EFBIG;
+		goto out;
+	}
+	*retblocks = size64 / blocksize;
+out:
 	close(fd);
-	*retblocks = (low + 1) / blocksize;
-	return 0;
+	return rc;
 }
 
 #endif /* WIN32 */
@@ -209,7 +312,7 @@ int main(int argc, char **argv)
 			"while calling ext2fs_get_device_size");
 		exit(1);
 	}
-	printf("Device %s has %d 1k blocks.\n", argv[1], blocks);
+	printf("Device %s has %u 1k blocks.\n", argv[1], blocks);
 	exit(0);
 }
 #endif
