@@ -22,12 +22,7 @@
 #include <errno.h>
 #endif
 
-#if EXT2_FLAT_INCLUDES
 #include "ext2_fs.h"
-#else
-#include <linux/ext2_fs.h>
-#endif
-
 #include "ext2fs.h"
 
 blk_t test_vec[] = { 8, 12, 24, 34, 43, 44, 100, 0 };
@@ -116,13 +111,13 @@ static void setup(void)
 		exit(1);
 	}
 	
-	retval = badblocks_list_create(&test_badblocks, 5);
+	retval = ext2fs_badblocks_list_create(&test_badblocks, 5);
 	if (retval) {
 		com_err("setup", retval, "while creating badblocks list");
 		exit(1);
 	}
 	for (i=0; test_vec[i]; i++) {
-		retval = badblocks_list_add(test_badblocks, test_vec[i]);
+		retval = ext2fs_badblocks_list_add(test_badblocks, test_vec[i]);
 		if (retval) {
 			com_err("setup", retval,
 				"while adding test vector %d", i);
@@ -141,7 +136,7 @@ static void iterate(void)
 	struct ext2_inode inode;
 	ext2_inode_scan	scan;
 	errcode_t	retval;
-	ino_t		ino;
+	ext2_ino_t	ino;
 
 	retval = ext2fs_open_inode_scan(test_fs, 8, &scan);
 	if (retval) {
@@ -180,7 +175,7 @@ static void check_map(void)
 
 	for (i=0; test_vec[i]; i++) {
 		if (ext2fs_test_block_bitmap(touched_map, test_vec[i])) {
-			printf("Bad block was touched --- %d\n", test_vec[i]);
+			printf("Bad block was touched --- %u\n", test_vec[i]);
 			failed++;
 			first_no_comma = 1;
 		}
@@ -204,7 +199,7 @@ static void check_map(void)
 				first = 0;
 			else
 				printf(", ");
-			printf("%d", i);
+			printf("%u", i);
 		}
 	}
 	printf("\n");
