@@ -244,21 +244,21 @@ static int process_file_block(ext2_filsys fs EXT2FS_ATTR((unused)),
 
 static void mark_table_blocks(ext2_filsys fs)
 {
-	blk_t	block, b;
+	blk_t	first_block, b;
 	unsigned int	i,j;
 	
-	block = fs->super->s_first_data_block;
+	first_block = fs->super->s_first_data_block;
 	/*
 	 * Mark primary superblock
 	 */
-	ext2fs_mark_block_bitmap(meta_block_map, block);
+	ext2fs_mark_block_bitmap(meta_block_map, first_block);
 			
 	/*
 	 * Mark the primary superblock descriptors
 	 */
 	for (j = 0; j < fs->desc_blocks; j++) {
 		ext2fs_mark_block_bitmap(meta_block_map,
-			 ext2fs_descriptor_block_loc(fs, block, j));
+			 ext2fs_descriptor_block_loc(fs, first_block, j));
 	}
 
 	for (i = 0; i < fs->group_desc_count; i++) {
@@ -287,7 +287,6 @@ static void mark_table_blocks(ext2_filsys fs)
 			ext2fs_mark_block_bitmap(meta_block_map,
 				 fs->group_desc[i].bg_inode_bitmap);
 		}
-		block += fs->super->s_blocks_per_group;
 	}
 }
 
@@ -418,7 +417,7 @@ static void output_meta_data_blocks(ext2_filsys fs, int fd)
 			retval = io_channel_read_blk(fs->io, blk, 1, buf);
 			if (retval) {
 				com_err(program_name, retval,
-					"error reading block %d", blk);
+					"error reading block %u", blk);
 			}
 			if (scramble_block_map && 
 			    ext2fs_test_block_bitmap(scramble_block_map, blk))
@@ -442,6 +441,8 @@ static void output_meta_data_blocks(ext2_filsys fs, int fd)
 		}
 	}
 	write_block(fd, zero_buf, sparse, 1, -1);
+	free(zero_buf);
+	free(buf);
 }
 
 static void write_raw_image_file(ext2_filsys fs, int fd, int scramble_flag)
@@ -517,7 +518,7 @@ static void write_raw_image_file(ext2_filsys fs, int fd, int scramble_flag)
 				       block_buf, process_dir_block, &pb);
 			if (retval) {
 				com_err(program_name, retval,
-					"while iterating over inode %d", 
+					"while iterating over inode %u",
 					ino);
 				exit(1);
 			}
@@ -530,7 +531,7 @@ static void write_raw_image_file(ext2_filsys fs, int fd, int scramble_flag)
 				       process_file_block, &pb);
 				if (retval) {
 					com_err(program_name, retval,
-					"while iterating over %d", ino);
+					"while iterating over inode %u", ino);
 					exit(1);
 				}
 			}
@@ -538,6 +539,7 @@ static void write_raw_image_file(ext2_filsys fs, int fd, int scramble_flag)
 	}
 	use_inode_shortcuts(fs, 0);
 	output_meta_data_blocks(fs, fd);
+	free(block_buf);
 }
 
 static void install_image(char *device, char *image_fn, int raw_flag)
@@ -629,7 +631,7 @@ int main (int argc, char ** argv)
 		 E2FSPROGS_DATE);
 	if (argc && *argv)
 		program_name = *argv;
-	initialize_ext2_error_table();
+	add_error_table(&et_ext2_error_table);
 	while ((c = getopt (argc, argv, "rsI")) != EOF)
 		switch (c) {
 		case 'r':
@@ -684,5 +686,6 @@ int main (int argc, char ** argv)
 		write_image_file(fs, fd);
 
 	ext2fs_close (fs);
+	remove_error_table(&et_ext2_error_table);
 	exit (0);
 }
