@@ -8,8 +8,6 @@
  *
  * Written by Theodore Ts'o, <tytso@mit.edu>
  * 
- * Usage:	fsck [-ACVRNTM] [-s] [-t fstype] [fs-options] device
- * 
  * Miquel van Smoorenburg (miquels@drinkel.ow.org) 20-Oct-1994:
  *   o Changed -t fstype to behave like with mount when -A (all file
  *     systems) or -M (like mount) is specified.
@@ -18,7 +16,8 @@
  *     can be added without changing this front-end.
  *   o -R flag skip root file system.
  *
- * Copyright (C) 1993, 1994, 1995, 1996, 1997, 1998, 1999 Theodore Ts'o.
+ * Copyright (C) 1993, 1994, 1995, 1996, 1997, 1998, 1999, 2000, 
+ * 	2001, 2002, 2003, 2004, 2005 by  Theodore Ts'o.
  *
  * %Begin-Header%
  * This file may be redistributed under the terms of the GNU Public
@@ -53,6 +52,9 @@
 #if HAVE_MALLOC_H
 #include <malloc.h>
 #endif
+#ifdef HAVE_SIGNAL_H
+#include <signal.h>
+#endif
 
 #include "../version.h"
 #include "nls-enable.h"
@@ -104,6 +106,7 @@ int like_mount = 0;
 int notitle = 0;
 int parallel_root = 0;
 int progress = 0;
+int progress_fd = 0;
 int force_all_parallel = 0;
 int num_running = 0;
 int max_running = 0;
@@ -129,6 +132,18 @@ static char *string_copy(const char *s)
 	return ret;
 }
 
+static int string_to_int(const char *s)
+{
+	long l;
+	char *p;
+
+	l = strtol(s, &p, 0);
+	if (*p || l == LONG_MIN || l == LONG_MAX || l < 0 || l > INT_MAX)
+		return -1;
+	else
+		return (int) l;
+}
+
 static int ignore(struct fs_info *);
 
 static char *skip_over_blank(char *cp)
@@ -353,10 +368,10 @@ static void load_fs_info(const char *filename)
 	fclose(f);
 	
 	if (old_fstab) {
-		fprintf(stderr, _("\007\007\007"
+		fputs(_("\007\007\007"
 		"WARNING: Your /etc/fstab does not contain the fsck passno\n"
 		"	field.  I will kludge around things for you, but you\n"
-		"	should fix your /etc/fstab file as soon as you can.\n\n"));
+		"	should fix your /etc/fstab file as soon as you can.\n\n"), stderr);
 		
 		for (fs = filesys_info; fs; fs = fs->next) {
 			fs->passno = 1;
@@ -439,10 +454,12 @@ static int execute(const char *type, const char *device, const char *mntpt,
 	for (i=0; i <num_args; i++)
 		argv[argc++] = string_copy(args[i]);
 
-	if (progress & !progress_active()) {
+	if (progress && !progress_active()) {
 		if ((strcmp(type, "ext2") == 0) ||
 		    (strcmp(type, "ext3") == 0)) {
-			argv[argc++] = string_copy("-C0");
+			char tmp[80];
+			snprintf(tmp, 80, "-C%d", progress_fd);
+			argv[argc++] = string_copy(tmp);
 			inst->flags |= FLAG_PROGRESS;
 		}
 	}
@@ -453,6 +470,7 @@ static int execute(const char *type, const char *device, const char *mntpt,
 	s = find_fsck(prog);
 	if (s == NULL) {
 		fprintf(stderr, _("fsck: %s: not found\n"), prog);
+		free(inst);
 		return ENOENT;
 	}
 
@@ -469,12 +487,14 @@ static int execute(const char *type, const char *device, const char *mntpt,
 		pid = -1;
 	else if ((pid = fork()) < 0) {
 		perror("fork");
+		free(inst);
 		return errno;
 	} else if (pid == 0) {
 		if (!interactive)
 			close(0);
 		(void) execv(s, argv);
 		perror(argv[0]);
+		free(inst);
 		exit(EXIT_ERROR);
 	}
 
@@ -729,8 +749,8 @@ static void compile_fs_type(char *fs_type, struct fs_type_compile *cmp)
 	cmp->list = malloc(num * sizeof(char *));
 	cmp->type = malloc(num * sizeof(int));
 	if (!cmp->list || !cmp->type) {
-		fprintf(stderr, _("Couldn't allocate memory for "
-				  "filesystem types\n"));
+		fputs(_("Couldn't allocate memory for filesystem types\n"), 
+		      stderr);
 		exit(EXIT_ERROR);
 	}
 	memset(cmp->list, 0, num * sizeof(char *));
@@ -766,7 +786,7 @@ static void compile_fs_type(char *fs_type, struct fs_type_compile *cmp)
 			}
 			if ((negate && !cmp->negate) ||
 			    (!negate && cmp->negate)) {
-				fprintf(stderr, _(fs_type_syntax_error));
+				fputs(_(fs_type_syntax_error), stderr);
 				exit(EXIT_USAGE);
 			}
 		}
@@ -925,7 +945,7 @@ static int check_all(NOARGS)
 	int pass_done;
 
 	if (verbose)
-		printf(_("Checking all file systems.\n"));
+		fputs(_("Checking all file systems.\n"), stdout);
 
 	/*
 	 * Do an initial scan over the filesystem; mark filesystems
@@ -1031,13 +1051,12 @@ static int check_all(NOARGS)
 
 static void usage(NOARGS)
 {
-	fprintf(stderr,
-		_("Usage: fsck [-ACNPRTV] [-t fstype] [fs-options] [filesys ...]\n"));
+	fputs(_("Usage: fsck [-ANPRTV] [ -C [ fd ] ] [-t fstype] [fs-options] [filesys ...]\n"), stderr);
 	exit(EXIT_USAGE);
 }
 
 #ifdef HAVE_SIGNAL_H
-static void signal_cancel(int sig)
+static void signal_cancel(int sig FSCK_ATTR((unused)))
 {
 	cancel_requested++;
 }
@@ -1125,6 +1144,22 @@ static void PRS(int argc, char *argv[])
 				break;
 			case 'C':
 				progress++;
+				if (arg[j+1]) {
+					progress_fd = string_to_int(arg+j+1);
+					if (progress_fd < 0)
+						progress_fd = 0;
+					else
+						goto next_arg;
+				} else if ((i+1) < argc && 
+					   !strncmp(argv[i+1], "-", 1) == 0) {
+					progress_fd = string_to_int(argv[i]);
+					if (progress_fd < 0)
+						progress_fd = 0;
+					else {
+						goto next_arg;
+						i++;
+					}
+				}
 				break;
 			case 'V':
 				verbose++;
@@ -1148,6 +1183,7 @@ static void PRS(int argc, char *argv[])
 				serialize++;
 				break;
 			case 't':
+				tmp = 0;
 				if (fstype)
 					usage();
 				if (arg[j+1])
@@ -1222,6 +1258,10 @@ int main(int argc, char *argv[])
 	if (oldpath) {
 		fsck_path = malloc (strlen (fsck_prefix_path) + 1 +
 				    strlen (oldpath) + 1);
+		if (!fsck_path) {
+			fprintf(stderr, "%s: Unable to allocate memory for fsck_path\n", progname);
+			exit(EXIT_ERROR);
+		}
 		strcpy (fsck_path, fsck_prefix_path);
 		strcat (fsck_path, ":");
 		strcat (fsck_path, oldpath);
