@@ -595,7 +595,7 @@ static void parse_tune2fs_options(int argc, char **argv)
 					tmp++;
 					break;
 				}
-				if (*tmp || interval > (365 * 86400)) {
+				if (*tmp) {
 					com_err (program_name, 0,
 						_("bad interval - %s"), optarg);
 					usage();
@@ -755,7 +755,7 @@ int main (int argc, char ** argv)
 #endif
 	if (argc && *argv)
 		program_name = *argv;
-	initialize_ext2_error_table();
+	add_error_table(&et_ext2_error_table);
 
 	if (strcmp(get_progname(argv[0]), "findfs") == 0)
 		do_findfs(argc, argv);
@@ -780,10 +780,12 @@ int main (int argc, char ** argv)
 		exit(1);
 	}
 	sb = fs->super;
+	fs->flags &= ~EXT2_FLAG_MASTER_SB_ONLY;
 	if (print_label) {
 		/* For e2label emulation */
 		printf("%.*s\n", (int) sizeof(sb->s_volume_name),
 		       sb->s_volume_name);
+		remove_error_table(&et_ext2_error_table);
 		exit(0);
 	}
 	retval = ext2fs_check_if_mounted(device_name, &mount_flags);
@@ -823,7 +825,8 @@ int main (int argc, char ** argv)
 		printf (_("Setting interval between checks to %lu seconds\n"), interval);
 	}
 	if (m_flag) {
-		sb->s_r_blocks_count = sb->s_blocks_count * reserved_ratio /100;
+		sb->s_r_blocks_count = e2p_percent(reserved_ratio,
+						   sb->s_blocks_count);
 		ext2fs_mark_super_dirty(fs);
 		printf (_("Setting reserved blocks percentage to %g%% (%u blocks)\n"),
 			reserved_ratio, sb->s_r_blocks_count);
@@ -919,5 +922,6 @@ int main (int argc, char ** argv)
 
 	if (l_flag)
 		list_super (sb);
+	remove_error_table(&et_ext2_error_table);
 	return (ext2fs_close (fs) ? 1 : 0);
 }
